import binascii

from .constants import *
from .util import *
from .communication import Communication
from TauLidarCommon.d3 import FrameBuilder
from TauLidarCommon.color import ColorMode, Color
from .info import CameraInfo
from TauLidarCommon.frame import FrameType, Frame

class Camera :
    '''
    ToF camera class, access point to the ToF camera.

    TauLidarCamera library depends on the TauLidarCommon library.

    If you install TauLidarCamera using pip (pip install TauLidarCamera), TauLidarCommon library will be automaticall installed. 
    
    To initiate an instance of Camera object, call static method of Camera.open(port=None).
    '''

    _frameBuilder = FrameBuilder()

    def __init__(self):
        self._comm = Communication()

    @staticmethod
    def open(port=None):
        '''
        open communication to the camera via serial port.

        By default, Camera will connect to the first available ToF device. 
        Alternatively can specify serial port '/dev/ttyACM0': 
        
            Camera.open('/dev/ttyACM0')

        Parameters
        ----------
        port : str, optional
            serial port the camera connected. if there is no port given, the first available camera will be used.

        Returns
        ----------
        Camera
            An instance of Camera object which connected to the ToF camera.

        Raises
        ----------
        Exception
            Error to open the camera.
        '''

        camera = Camera()
        return camera._open(port)

    @staticmethod
    def scan():
        '''
        scan for available Tau LiDAR cameras


        Parameters
        ----------
        None

        Returns
        ----------
        ports
            A list of serial port device name that represents Tau cameras

        '''

        camera = Camera()
        return camera._comm.scan()

    def _open(self, port=None):

        if self._comm._ser.is_open :
            raise Exception("Camera is already opened!")

        self._comm.open(port)

        if self._comm._ser.is_open: 
            self.setDefaultParameters()

        return self

    def info(self):
        '''
        get camera information, such as model, firmware version, uid, resolution and serial port name.

        Returns
        ----------
        CameraInfo
            An instance of CameraInfo contains model, firmware version, uid, resolution and serial port name.
        '''

        chipType, chipVersion = self.getIdentification()
        model = "%d.%d" % (chipType, chipVersion)

        major, minor = self.getFirmwareRelease()
        firmware = "%d.%d" % (major, minor)

        waferId, chipId = self.getChipInformation()
        uid = "%02X%04X" % (waferId, chipId)

        return CameraInfo(model, firmware, uid, "160x60",  self._comm._ser.port)

    def close(self) :
        '''
        Close communication to the camera.
        '''

        self._comm.close()

    def setDefaultParameters(self) :
        '''
        Convenient method to set default parameters to control the ToF camera. You may set those parameters
        in your own application.
        '''
        self.setModulationFrequency(VALUE_20MHZ) ## frequency: 20MHZ
        self.setModulationChannel(0)             ## autoChannelEnabled: 0, channel: 0
        self.setMode(0)                          ## Mode 0, wide fov
        self.setHdr(0)                           ## HDR off
        self.setIntegrationTime3d(0, 800)        ## set integration time 0: 1000
        self.setIntegrationTimeGrayscale(5000)   ## set integration time grayscale: 20000
        self.setMinimalAmplitude(0, 60)          ## set minimal amplitude 0: 80
        self.setOffset(0)                        ## set distance offset: 0
        self.setRoi(0, 0, 159, 59)               ## set ROI to max width and height

    def setModulationFrequency(self, frequency):
        '''
        Set Modulation Frequency, for wide view image, it has to be set to VALUE_20MHZ.
        '''
        self._comm.setModulationFrequency(frequency)

    def setModulationChannel(self, channel):
        '''
        Set Modulation Channel, for multiple cameras, you may set them work in different channels, 
        however if you have only one, set it to 0.
        '''
        self._comm.setModulationChannel(0, channel)

    def setMode(self, mode):
        '''
        Set camera mode. Currently supports wide view iamge only, which is 0.
        '''
        self._comm.setMode(mode)

    def setHdr(self, hdr):
        '''
        Set HDR mode. Currently not supported.
        '''
        self._comm.setHdr(hdr)

    def setIntegrationTime3d(self, index, t):
        '''
        Set Integration Time. 
        Integration Time could be from 0 to 1000, depends on the surface reflectivity and distance of the target object, 
        for an object in about 2-3 meters, you may try set it to 800.
        '''
        self._comm.setIntegrationTime3d(index, t)

    def setIntegrationTimeGrayscale(self, t):
        '''
        Set Integration Time for Grayscale. 
        If you request distance/depth plus grayscale image, you need set approriate “integration time grayscale”, 
        maximum allowed 25000, for example, regular in-door environment, set it to 8000.
        '''
        self._comm.setIntegrationTimeGrayscale(t)

    def setMinimalAmplitude(self, index, t):
        '''
        Set MinimalAmplitude. 
        MinimalAmplitude is the noise level minimum threshold, 
        for an object in about 2-3 meters in in-door environment, you may try set it to 60.
        '''
        self._comm.setMinimalAmplitude(index, t)

    def setOffset(self, offset):
        '''
        Set distance offset. 
        Normally this set it to 0.
        '''
        self._comm.setOffset(offset)

    def setRoi(self, x0, y0, x1, y1):
        '''
        Set ROI. 
        Currently not supported.
        '''
        self._comm.setRoi(x0, y0, x1, y1)

    @staticmethod
    def setColorMode(colorMode):
        '''
        Set ColorMode. 
        To use distance or grayscle for point cloud color.
        '''
        Camera._frameBuilder.setColorMode(colorMode)

    @staticmethod
    def setRange(z1, z2):
        '''
        Set Range. 
        To use distance for point cloud color, set the distance range.
        '''
        Camera._frameBuilder.setRange(z1, z2)

    def getChipInformation(self) :
        '''
        request Chip Information.
        '''
        dataArray = self._comm.getChipInformation()
        waferId = getUint16LittleEndian(dataArray, INDEX_WAFER_ID - COMMAND_SIZE_HEADER)
        chipId  = getUint16LittleEndian(dataArray, INDEX_CHIP_ID  - COMMAND_SIZE_HEADER)

        return (waferId, chipId)

    def getIdentification(self) :
        '''
        request Identification.
        '''
        dataArray = self._comm.getIdentification()
        identificationValue = int(binascii.hexlify(dataArray), 16)
        chipType = (identificationValue & MASK_CHIP_TYPE_DEVICE) >> SHIFT_CHIP_TYPE_DEVICE
        chipVersion = (identificationValue & MASK_VERSION) >> SHIFT_VERSION

        return (chipType, chipVersion)

    def getFirmwareRelease(self) :
        '''
        request Firmware Release.
        '''
        dataArray = self._comm.getFirmwareRelease()
        firmware = getUint32LittleEndian(dataArray, 0)
        major = getValueMsb(firmware)
        minor = getValueLsb(firmware)
        return (major, minor)

    def readFrameRawData(self, frameType) :
        '''
        To request raw data of a frame from camera.

        You may use readFrameRawData(frameType) only when you consider a two-step read-and-compose a Frame object, which readFrameRawData(frameType)
        to obtain a frame raw data, and then use FrameBuilder.composeFrame(dataArray, frameType) to create an Frame object, normally in
        a separate thread, to have a better frame rate.

        You may simply call readFrame(frameType) to obtain a Frame object without using this method if the frame rate is not critical.

        FrameType.DISTANCE: depth data only

        FrameType.DISTANCE_GRAYSCALE: depth data plus grayscale

        FrameType.DISTANCE_AMPLITUDE: depth data plus amplitude

        Returns
        ----------
        bytearray
            Length of the raw data will be different depends on the frameType specified:

        FrameType.DISTANCE: 2 bytes 32 float distance for each data point, 160 (image width) x 160 (image height) x 2;

        FrameType.DISTANCE_GRAYSCALE: 2 bytes 32 float distance and 1 byte unit 8 grayscale for each data point, 160 (image width) x 160 (image height) x (2 + 1);

        FrameType.DISTANCE_AMPLITUDE: 2 bytes 32 float distance and 2 bytes 32 float amplitude for each data point, 160 (image width) x 160 (image height) x (2 + 2);
        '''
        
        if frameType == FrameType.DISTANCE_GRAYSCALE:
            dataArray = self._comm.getDistanceGrayscale()
            return dataArray[TOF_635_IMAGE_HEADER_SIZE : len(dataArray)]

        elif frameType == FrameType.DISTANCE_AMPLITUDE:
            dataArray = self._comm.getDistanceAmplitude()
            return dataArray[TOF_635_IMAGE_HEADER_SIZE : len(dataArray)]

        elif frameType == FrameType.DISTANCE:
            dataArray = self._comm.getDistance()
            return dataArray[TOF_635_IMAGE_HEADER_SIZE : len(dataArray)]

    @staticmethod
    def composeFrame(dataArray, frameType) :
        '''
        Convenient method to compose Frame using raw bytearray data.

        To compose a Frame object, you need to specify which FrameType of the raw data, which may be:

        FrameType.DISTANCE: distance / depth only

        FrameType.DISTANCE_GRAYSCALE: distance / depth plus grayscale

        FrameType.DISTANCE_AMPLITUDE: distance / depth plus amplitude

        Returns
        ----------
        Frame
            An instance of Frame.

            The members of the Frame object will be populated:
            
                data_depth,
                data_depth_rgb,
                points_3d

            In addition, depends on the FrameType requested, additional member may be populated:

            FrameType.DISTANCE_GRAYSCALE: data_grayscale.

            FrameType.DISTANCE_AMPLITUDE: data_amplitude.
        '''
        
        return Camera._frameBuilder.composeFrame(dataArray, frameType)

    def readFrame(self, frameType=FrameType.DISTANCE_GRAYSCALE) :
        '''
        A convenient method to directly get a new Frame object.

        Specifying a frameType to get a frame with distance / depth only, or distance / depth plus grayscale or amplitude:

        FrameType.DISTANCE: distance / depth only

        FrameType.DISTANCE_GRAYSCALE: distance / depth plus grayscale

        FrameType.DISTANCE_AMPLITUDE: distance / depth plus amplitude
        
        This method is an expensive call, alternatively you may use readFrameRawData and compose Frame from a separate thread in your application
        to get better frame rate.

        For example, for an application which the frame rate is not critical, a simple call of readFrame to get an instance of Frame object:

        frame = camera.readFrame(self, frameType=FrameType.DISTANCE)

        If your application require a fast frame rate, it might be a good idea to request frame raw data and compose Frame object in different threads:
        
        The local thread to request a frame raw data:

        dataArray = self.readFrameRawData(frameType)

        In another thread to compose the Frame object from dataArray, and the local thread can continue to request next frame raw data:

        _frameBuilder.composeFrame(dataArray, frameType)

        Returns
        ----------
        Frame
            An instance of Frame.

        The members of the Frame object will be populated:
            
                data_depth,
                data_depth_rgb,
                points_3d

            In addition, depends on the FrameType requested, additional member may be populated:

            FrameType.DISTANCE_GRAYSCALE: data_grayscale.

            FrameType.DISTANCE_AMPLITUDE: data_amplitude.

        Example of converting data_depth_rgb to OpenCV Mat object:
                
            mat_depth_rgb = np.frombuffer(frame.data_depth_rgb, dtype=np.uint16, count=-1, offset=0).reshape(frame.height, frame.width, 3)
            mat_depth_rgb = mat_depth_rgb.astype(np.uint8)

        Example of converting data_grayscale to OpenCV Mat object:
                
            mat_grayscale = np.frombuffer(frame.data_grayscale, dtype=np.uint16, count=-1, offset=0).reshape(frame.height, frame.width)
            mat_grayscale = mat_grayscale.astype(np.uint8)

        Example of converting data_amplitude to OpenCV Mat object:

            mat_amplitude = np.frombuffer(frame.data_amplitude, dtype=np.float32, count=-1, offset=0).reshape(frame.height, frame.width)
            mat_amplitude = mat_amplitude.astype(np.uint8)
        '''
        
        dataArray = self.readFrameRawData(frameType)

        return Camera.composeFrame(dataArray, frameType)




class CameraInfo :
    '''
    Camera Information.
    
    Attributes
    ----------
    model: 
        Camera model.
    firmware: 
        Firmware version.
    uid: 
        Camera identification.
    resolution:
        Camera resolution
    port: 
        Serial port
    '''
    def __init__(self, _model, _firmware, _uid, _resolution, _port) :
        self.model      = _model
        self.firmware   = _firmware
        self.uid        = _uid
        self.resolution = _resolution
        self.port       = _port



import serial
import serial.tools.list_ports
from time import sleep
import binascii

from .constants import *
from .crc import *
from .util import *

class Communication:
    '''
    Communication to the ToF sensor via serial port.
    '''

    def __init__(self):
        self._ser = serial.Serial()
        self._ser.baudrate = 4000000
        self._ser.timeout = 0.001
        self._ser.parity = serial.PARITY_NONE
        self._ser.stopbits = serial.STOPBITS_ONE
        self._ser.bytesize = serial.EIGHTBITS


    def scan(self):
        deviceList = []
        print('Looking for connected Tau LiDAR Camera hardware ...')

        ports = list(serial.tools.list_ports.comports())
        for port, description, address in ports:
            self._ser.port = port
            try:
                self._ser.open()
            except:
                pass
            if self._ser.is_open:
                try:
                    ## Verify if it is valid device
                    dataArray = self.getIdentification()
                    identificationValue = int(binascii.hexlify(dataArray), 16)
                    chipType = (identificationValue & MASK_CHIP_TYPE_DEVICE) >> SHIFT_CHIP_TYPE_DEVICE
                    chipVersion = (identificationValue & MASK_VERSION) >> SHIFT_VERSION

                    nChipType = int(chipType)
                    nChipVersion = int(chipVersion)

                    if (nChipType >= 4 and nChipVersion >= 0):
                        deviceList.append(port)
                except:
                    pass
                self._ser.close()

        if len(deviceList) == 0:
            print("No Tau Camera devices found, please check: \n1. If Tau Camera is connected; \n2. If current user has permission to access all serial ports.")

        return deviceList

    def open(self, port):
        '''
        auto detect sensor at serial port and open communication to the sensor if port is not given,
        or open communication to specific serial port.

        Parameters
        ----------
        port: str
            serial port the sensor connected.
        '''
        if self._ser.is_open :
            raise Exception("Serial port is already opened!")

        message = ""
        if port is None:
            print('Looking for connected Tau LiDAR Camera hardware ...')
            ports = list(serial.tools.list_ports.comports())
            connected = False
            for port_no, description, address in ports:
                try:
                    self._ser.port = port_no
                    self._ser.open()

                    ## Verify if it is valid device
                    dataArray = self.getIdentification()
                    identificationValue = int(binascii.hexlify(dataArray), 16)
                    chipType = (identificationValue & MASK_CHIP_TYPE_DEVICE) >> SHIFT_CHIP_TYPE_DEVICE
                    chipVersion = (identificationValue & MASK_VERSION) >> SHIFT_VERSION

                    nChipType = int(chipType)
                    nChipVersion = int(chipVersion)

                    if (nChipType >= 4 and nChipVersion >= 0):
                        connected = True
                        break
                except:
                    self._ser.close()

            if not connected:
                message = "No Tau Camera found, please check: \n1. If Tau Camera is connected; \n2. If current user has permission to access all serial ports."
                raise Exception(message)
        else:
            self._ser.port = port
            try:
                self._ser.open()
            except:
                message = "Failed connecting to the serial port %s, please check: \n1. If Tau Camera is connected; \n2. If current user has permission to access the port" % port
                raise Exception(message)

        return self

    def close(self):
        '''
        close communication to the sensor.
        '''
        if self._ser.is_open :
            self._ser.close()

    def _write(self, data):

        if not self._ser.is_open :
            print("ERROR: serial port is not open!")
            return

        data[0] = COMMAND_START_MARK
        checksum = calculateChecksum(data, (COMMAND_SIZE_TOTAL - 4))

        setUint32LittleEndian(data, (COMMAND_SIZE_TOTAL - 4), checksum)

        ################################
        ##debeg_str= "SEND COMMAND: "
        ##for i in range(14) :
        ##    buf = "%x" % data[i]
        ##    debeg_str += buf
        ##print(debeg_str)
        ################################

        return self._ser.write(data)

    def _read(self, size):

        if not self._ser.is_open :
            print("ERROR: serial port is not open!")
            return

        size_with_command = size + 8

        array = bytearray(0)
        count = 0
        while (len(array) < size_with_command) :
            len_pending = size_with_command - len(array)
            in_waiting = self._ser.in_waiting
            sz = max(len_pending, self._ser.in_waiting)
            
            buf = self._ser.read(sz)
            a = bytearray(buf)
            array.extend(a)

            count += 1
            if count > 100 : break

        return self._processData(array, size)

    def _processData(self, array, size):

        data_length = len(array)

        if data_length == 0 :
            # print("      DATA ERROR - length is 0")
            #return False
            return (-1, bytearray(0))

        ## Check for the start mark
        if (array[0] != DATA_START_MARK) :
            #print("      DATA ERROR - the start byte is not DATA_START_MARK")
            #return True
            return (-1, bytearray(0))

        if data_length < COMMAND_SIZE_HEADER :
            #print("      DATA ERROR - length shorter than COMMAND_SIZE_HEADER")
            #return True
            return (-1, bytearray(0))

        ## Cancel here if not enough bytes received
        if data_length < COMMAND_SIZE_OVERHEAD :
            #print("      DATA ERROR - length shorter than COMMAND_SIZE_OVERHEAD")
            #return True
            return (-1, bytearray(0))

        ## Get the expexted size
        expectedSize = self._getExpextedSize(array)

        #if checksumIsCorrect(array, expectedSize) : ## calculate checksum for a frame data size 28880 is too time consumming, turn off checking for now
        type = self._getType(array)
        dataArray = array
        dataArray[0:COMMAND_SIZE_HEADER] = []

        ## Remove checksum at the end
        dataArray[expectedSize:expectedSize + COMMAND_SIZE_CHECKSUM] = []

        ## Remove the rest at the end
        dataArray[expectedSize:len(dataArray)] = []

        data_length = len(dataArray)

        if (data_length < size):
            print("Data error, actual size: %d, expected size: %d" % (data_length, size))

        return (type, dataArray)

    def _sendCommand(self, data, size):

        if self._write(data) > 0:
            return self._read(size)

        return None

    def _sendCommandWithoutData(self, command, size):

        data = bytearray(14)
        data[1] = command
        return self._sendCommand(data, size)

    def _sendCommandSingleByte(self, command, payload, size):
        data = bytearray(COMMAND_SIZE_TOTAL)

        ## Add the command
        data[COMMAND_INDEX_COMMAND] = command

        ## Add the single byte at the first position
        data[COMMAND_INDEX_DATA] = payload

        return self._sendCommand(data, size)

    def _sendCommand2xByte(self, command, payload0, payload1, type):
        data = bytearray(COMMAND_SIZE_TOTAL)

        ## Add the command
        data[COMMAND_INDEX_COMMAND] = command

        ## Add the first byte
        data[COMMAND_INDEX_DATA] = payload0

        ## Add the second byte
        data[COMMAND_INDEX_DATA + 1] = payload1

        return self._sendCommand(data, type)

    def _sendCommandUint16(self, command, payload, size):
        data = bytearray(COMMAND_SIZE_TOTAL)

        ## Add the command
        data[COMMAND_INDEX_COMMAND] = command

        ## Add the payload
        setUint16LittleEndian(data, COMMAND_INDEX_DATA, payload)

        return self._sendCommand(data, size)

    def _sendCommandInt16(self, command, payload, size):
        data = bytearray(COMMAND_SIZE_TOTAL)

        ## Add the command
        data[COMMAND_INDEX_COMMAND] = command

        ## Add the payload
        setInt16LittleEndian(data, COMMAND_INDEX_DATA, payload)

        return self._sendCommand(data, size)

    def _getExpextedSize(self, array):
        return getUint16LittleEndian(array, DATA_INDEX_LENGTH)

    def _getType(self, array) :
        return array[DATA_INDEX_TYPE]

    def setModulationFrequency(self, frequency):
        '''
        set Modulation Frequency.

        Parameters
        ----------
        frequency
            VALUE_10MHZ (10mHz) or VALUE_20MHZ (20mHz).
        '''
        return self._sendCommandSingleByte(COMMAND_SET_MODULATION_FREQUENCY, frequency, 0) ## expected data length: 0

    def setModulationChannel(self, autoChannelEnabled, channel):
        '''
        set Modulation Channel.

        Parameters
        ----------
        autoChannelEnabled
            on or off the auto Channel
        channel
            0 - 10.
        '''
        return self._sendCommand2xByte(COMMAND_SET_MODULATION_CHANNEL, autoChannelEnabled, channel, DATA_ACK)

    def setMode(self, mode):
        '''
        set Mode.

        Parameters
        ----------
        mode
            0 - wide FOV
        '''
        return self._sendCommandSingleByte(COMMAND_SET_MODE, mode, 0) ## expected data length: 0

    def setHdr(self, hdr):
        '''
        set HDR.

        Parameters
        ----------
        hdr
            0 - HDR Off
        '''
        self._hdr = hdr

        return self._sendCommandSingleByte(COMMAND_SET_HDR,  hdr, 0)  ## expected data length: 0

    def setIntegrationTimeGrayscale(self, integrationTime):
        '''
        set Integration Time Grayscale.

        Parameters
        ----------
        integrationTime
            0 - 50000
        '''
        return self._sendCommandUint16(COMMAND_SET_INTEGRATION_TIME_GRAYSCALE, integrationTime, 0)

    def setOffset(self, offset):
        '''
        set Offset.

        Parameters
        ----------
        offset
            distance offset
        '''
        return self._sendCommandInt16(COMMAND_SET_OFFSET, offset, 0)  ## expected data length: 0

    def setIntegrationTime3d(self, index, integrationTime):
        '''
        set set Integration Time 3d.

        Parameters
        ----------
        index
            integrationTime index
        integrationTime
            integrationTime
        '''
        data = bytearray(COMMAND_SIZE_TOTAL)

        ## Add the command
        data[COMMAND_INDEX_COMMAND] = COMMAND_SET_INTEGRATION_TIME_3D

        ## Add the index
        data[INDEX_INDEX_3D] = index

        ## Add the time
        setUint16LittleEndian(data, INDEX_INTEGRATION_TIME_3D, integrationTime)

        return self._sendCommand(data, 0)

    def setMinimalAmplitude(self, index, amplitude):
        '''
        set Minimal Amplitude.

        Parameters
        ----------
        index
            MinimalAmplitude index
        amplitude
            amplitude
        '''
        data = bytearray(COMMAND_SIZE_TOTAL)

        ## Add the command
        data[COMMAND_INDEX_COMMAND] = COMMAND_SET_MINIMAL_AMPLITUDE

        ## Add the index
        data[INDEX_INDEX_AMPLITUDE] = index

        ## Add the amplitude
        setUint16LittleEndian(data, INDEX_AMPLITUDE, amplitude)

        return self._sendCommand(data, 0)

    def setRoi(self, xMin, yMin, xMax, yMax):
        '''
        set Roi.

        Parameters
        ----------
        xMin
            xMin
        yMin
            yMin
        xMax
            xMax
        yMax
            yMax
        '''
        data = bytearray(COMMAND_SIZE_TOTAL + 4 * 2) ## 4 two-bytes data (uint16_t in c++)

        ## Add the command
        data[COMMAND_INDEX_COMMAND] = COMMAND_SET_ROI

        ## xMin
        setUint16LittleEndian(data, INDEX_ROI_X_MIN, xMin)

        ## yMin
        setUint16LittleEndian(data, INDEX_ROI_Y_MIN, yMin)

        ## xMax
        setUint16LittleEndian(data, INDEX_ROI_X_MAX, xMax)

        ## yMax
        setUint16LittleEndian(data, INDEX_ROI_Y_MAX, yMax)

        self._xMin = xMin
        self._yMin = yMin
        self._xMax = xMax
        self._yMax = yMax

        return self._sendCommand(data, 0)

    def getChipInformation(self):
        '''
        get Chip Information.
        '''
        type, data = self._sendCommandWithoutData(COMMAND_GET_CHIP_INFORMATION, CHIP_INFORMATION_DATA_SIZE)
        return data

    def getIdentification(self):
        '''
        get getIdentification.
        '''
        type, data = self._sendCommandWithoutData(COMMAND_GET_IDENTIFICATION, IDENTIFICATION_DATA_SIZE)
        return data

    def getFirmwareRelease(self):
        '''
        get Firmware Release.
        '''
        type, data = self._sendCommandWithoutData(COMMAND_GET_FIRMWARE_RELEASE, FIRMWARE_RELEASE_DATA_SIZE)
        return data

    def getDistanceGrayscale(self):
        '''
        get Distance Grayscale image.
        '''
        dataSize = 0

        if self._hdr == HDR_SPATIAL : dataSize = 3 * (self._xMax - self._xMin + 1) * (self._yMax - self._yMin + 1) / 2 + 80 ## 16 bit distance/2 + 8 bit grayscale/2
        else :
            dataSize = 3 * (self._xMax - self._xMin + 1) * (self._yMax - self._yMin + 1) + 80 ## 16 bit distance + 8 bit grayscale;

        type, data = self._sendCommandSingleByte(COMMAND_GET_DISTANCE_GRAYSCALE, AUTO_REPEAT, dataSize)
        return data

    def getDistanceAmplitude(self):
        '''
        get Distance Amplitude image.
        '''
        dataSize = 0

        if self._hdr == HDR_SPATIAL : dataSize = 2 * (self._xMax - self._xMin + 1) * (self._yMax - self._yMin + 1) + 80 ## 16 bit distance/2 + 16 bit
        else :
            dataSize = 4 * (self._xMax - self._xMin + 1) * (self._yMax - self._yMin + 1) + 80 ## 16 bit distance + 16 bit amplitude

        type, data = self._sendCommandSingleByte(COMMAND_GET_DISTANCE_AMPLITUDE, AUTO_REPEAT, dataSize)
        return data

    def getDistance(self):
        '''
        get Distance only.
        '''
        dataSize = 0

        if self._hdr == HDR_SPATIAL : dataSize = (self._xMax - self._xMin + 1) * (self._yMax - self._yMin + 1) + 80 ## 16 bit distance
        else :
            dataSize = 2 * (self._xMax - self._xMin + 1) * (self._yMax - self._yMin + 1) + 80 ## 16 bit distance

        type, data = self._sendCommandSingleByte(COMMAND_GET_DISTANCE, AUTO_REPEAT, dataSize)
        return data



from .constants import *
from .util import *

POLYNOM = 0x04C11DB7
CRC_INIT_VALUE = 0xFFFFFFFF
XOR_VALUE = 0x00000000

def _calcCrc32Uint32(crc, data) :
    crc = crc ^ data
    for i in range(32) :
        if crc & 0x80000000 :
            crc = (crc << 1) ^ POLYNOM
        else :
            crc = crc << 1
    return(crc)

def _calcCrc32_32(data, size) :
    crc = CRC_INIT_VALUE
    for i in range(size) :
        crc = _calcCrc32Uint32(crc, data[i])
    return crc ^ XOR_VALUE

def calculateChecksum(data, size) :
    '''
    Calculate Checksum.

    Parameters
    ----------
    data : bytearray
        bytearray data.
    size : int
        size.
    '''
    return _calcCrc32_32(data, size)

def checksumIsCorrect(array, expectedSize) :
    '''
    Verify if the checksum is correct.

    Parameters
    ----------
    array : bytearray
        bytearray data.
    expectedSize : int
        expected size.
    '''
    ## The received CRC is the one in the data
    receivedCrc = getUint32LittleEndian(array, SIZE_HEADER + expectedSize)

    ## The wanted CRC is the one calculated out of the payload
    wantedCrc = calculateChecksum(array, SIZE_HEADER + expectedSize)

    byte0 = wantedCrc & 0xFF
    byte1 = (wantedCrc >> 8) & 0xFF
    byte2 = (wantedCrc >> 16) & 0xFF
    byte3 = (wantedCrc >> 24) & 0xFF
    wantedCrc_value = (byte3 << 24) | (byte2 << 16) | (byte1 << 8) | byte0

    if receivedCrc == wantedCrc_value :
        return True

    print("Checksum ERROR!!!")

    return False



## Command format
COMMAND_SIZE_TOTAL    = 14                                  ## Cammand size total
COMMAND_SIZE_HEADER   = 4                                   ## Cammand header size
COMMAND_SIZE_CHECKSUM = 4                                   ## Cammand checksum size
COMMAND_SIZE_OVERHEAD = 8                                   ## COMMAND_SIZE_HEADER + COMMAND_SIZE_CHECKSUM
COMMAND_START_MARK    = 0xF5                                ## Command start marking
COMMAND_INDEX_COMMAND = 1                                   ## Cammand index
COMMAND_INDEX_DATA    = 2                                   ## Cammand payload data

## setup commands
COMMAND_SET_INTEGRATION_TIME_3D = 0x00                      ## Command to set the integration time for 3D operation
COMMAND_SET_INTEGRATION_TIME_GRAYSCALE = 0x01               ## Command to set the integration time for grayscale
COMMAND_SET_ROI = 0x02                                      ## Command to set the region of interest
COMMAND_SET_BINNING = 0x03                                  ## Command to set the binning
COMMAND_SET_MODE = 0x04                                     ## Command to set the mode
COMMAND_SET_MODULATION_FREQUENCY = 0x05                     ## Command to set the modulation frequency
COMMAND_SET_DLL_STEP = 0x06                                 ## Command to set the DLL step
COMMAND_SET_FILTER = 0x07                                   ## Command to set the filter parameters
COMMAND_SET_OFFSET = 0x08                                   ## Command to set the offset
COMMAND_SET_MINIMAL_AMPLITUDE = 0x09                        ## Command to set th minimal amplitude
COMMAND_SET_DCS_FILTER = 0x0A                               ## Command to set the DCS filter
COMMAND_SET_GAUSSIAN_FILTER = 0x0B                          ## Command to set the Gaussian filter
COMMAND_SET_FRAME_RATE = 0x0C                               ## Command to set/limit the frame rate
COMMAND_SET_HDR = 0x0D
COMMAND_SET_MODULATION_CHANNEL = 0x0E                       ## Command to set the modulation channel
COMMAND_SET_FILTER_SINGLE_SPOT = 0x0F                       ## Command to set the temporal filter for the single spot

## acquisition commands
COMMAND_GET_DISTANCE = 0x20                                 ## Command to request distance data
COMMAND_GET_AMPLITUDE = 0x21                                ## Command to request amplitude data
COMMAND_GET_DISTANCE_AMPLITUDE = 0x22                       ## Command to request distance and amplitude data
COMMAND_GET_DCS_DISTANCE_AMPLITUDE = 0x23                   ## Command to request distance, amplitude and DCS data at once
COMMAND_GET_GRAYSCALE = 0x24                                ## Command to request grayscale data
COMMAND_GET_DCS = 0x25                                      ## Command to request DCS data
COMMAND_SET_AUTO_ACQUISITION = 0x26                         ## Command to enable/disable the auto acquisition
COMMAND_GET_INTEGRATION_TIME_3D = 0x27                      ## Command to read the integration time. Important when using automatic mode
COMMAND_STOP_STREAM = 0x28                                  ## Command to stop the stream
COMMAND_GET_DISTANCE_GRAYSCALE = 0x29                       ## Command to request distance and grayscale

COMMAND_GET_IDENTIFICATION   = 0x47                         ## Command to identification
COMMAND_GET_CHIP_INFORMATION = 0x48                         ## Command to read the chip information
COMMAND_GET_FIRMWARE_RELEASE = 0x49                         ## Command to read the firmware release
COMMAND_GET_PRODUCTION_INFO  = 0x50                         ## Command to get the production info

## MODULATION
COMMAND_SET_MODULATION_FREQUENCY = 0x05
VALUE_10MHZ = 0 ## Value for 10MHz for command "COMMAND_SET_MODULATION_FREQUENCY"
VALUE_20MHZ = 1 ## Value for 20MHz for command "COMMAND_SET_MODULATION_FREQUENCY"

## 635 op mode
MODE_BEAM_A = 0                                             ## Normal operation with illumination beam A
MODE_BEAM_B_MANUAL = 1                                      ## Normal operation with illumination beam B (all settings by user, same as)
MODE_BEAM_B_RESULT = 2                                      ## Beam B with calibrated ROI, only one distance as result
MODE_BEAM_B_RESULT_DATA = 3                                 ## Beam B with calibrated ROI, one distance and the pixels as result
MODE_BEAM_AB_RESULT = 4                                     ## Beam A and B operating with calibrated ROI and only one distance as result
MODE_BEAM_AB_AUTO_RESULT = 5                                ## Beam A and B with automatic selection
MODE_BEAM_AB_INTERLEAVED_DATA = 6                           ## Beam A and B interleaved output

## Stream mode
SINGLE = 0                                                  ## Single frame mode
AUTO_REPEAT = 1                                             ## Auto repeat frame using same parameters
STREAM = 3                                                  ## Stream mode

## HDR
HDR_OFF = 0                                                 ## HDR off
HDR_SPATIAL = 1                                             ## Spatial HDR
HDR_TEMPORAL = 2                                            ## Temporal HDR

## IntegrationTime
INDEX_INDEX_3D = 2                                          ## Index of the integration time 3d index
INDEX_INTEGRATION_TIME_3D = 3                               ## Index of the integration time 3d

## AMPLITUDE
INDEX_INDEX_AMPLITUDE = 2                                   ## Index of the index
INDEX_AMPLITUDE = 3                                         ## Index of the minimal amplitude

## ROI
INDEX_ROI_X_MIN = 2                                         ## Index of ROI X MIN
INDEX_ROI_Y_MIN = 4                                         ## Index of ROI Y MIN
INDEX_ROI_X_MAX = 6                                         ## Index of ROI X MAX
INDEX_ROI_Y_MAX = 8                                         ## Index of ROI Y MAX

## Data format
DATA_START_MARK    = 0xFA                                   ## Data start marking
DATA_INDEX_LENGTH  = 2                                      ## Data length
DATA_INDEX_TYPE    = 1                                      ## Data type

## firmware returned data type
DATA_ACK = 0x00                                             ## Acknowledge from sensor to host
DATA_NACK = 0x01                                            ## Not acknowledge from sensor to host
DATA_IDENTIFICATION = 0x02                                  ## Identification to identify the device
DATA_DISTANCE = 0x03                                        ## Distance information
DATA_AMPLITUDE = 0x04                                       ## Amplitude information
DATA_DISTANCE_AMPLITUDE = 0x05                              ## Distance and amplitude information
DATA_GRAYSCALE = 0x06                                       ## Grayscale information
DATA_DCS = 0x07                                             ## DCS data
DATA_DCS_DISTANCE_AMPLITUDE = 0x08                          ## DCS, distance and amplitude all together
DATA_INTEGRATION_TIME = 0x09                                ## Integration time, answer to COMMAND_GET_INTEGRATION_TIME_3D
DATA_DISTANCE_GRAYSCALE = 0x0A                              ## Distance and grayscale data
DATA_LENS_CALIBRATION_DATA = 0xF7                           ## Lens calibration data
DATA_TRACE = 0xF8                                           ## Trace data
DATA_PRODUCTION_INFO = 0xF9                                 ## Production info
DATA_CALIBRATION_DATA = 0xFA                                ## Calibration data
DATA_REGISTER = 0xFB                                        ## Register data
DATA_TEMPERATURE = 0xFC                                     ## Temperature data
DATA_CHIP_INFORMATION = 0xFD                                ## Chip information data
DATA_FIRMWARE_RELEASE = 0xFE                                ## Firmware release
DATA_ERROR = 0xFF                                           ## Error number

## CHIP
MASK_CHIP_TYPE_DEVICE = 0x00FFFF00                          ## Chip information mask
SHIFT_CHIP_TYPE_DEVICE = 8                                  ## Chip information shift
CHIP_INFORMATION_DATA_SIZE = 4                              ## Chip information data size

## IDENTITY
DATA_IDENTIFICATION = 0x02
DATA_FIRMWARE_RELEASE = 0xFE
IDENTIFICATION_DATA_SIZE = 4                                ## Chip information data size
INDEX_WAFER_ID = 6
INDEX_CHIP_ID = 4

## Firmware release
FIRMWARE_RELEASE_DATA_SIZE = 4                              ## Chip information data size
MASK_VERSION = 0x000000FF
SHIFT_VERSION = 0

## TOF 635 image
TOF_635_IMAGE_HEADER_SIZE  = 80                            ## 635 IMAGE HEADER SIZE



def setUint32LittleEndian(buffer, index, value) :
    '''
    set Uint32 Little Endian.

    Parameters
    ----------
    buffer : bytearray
        data array.
    index : int
        starting index.
    value : int
        value to be set.
    '''
    buffer[index] = value & 0xFF
    buffer[index+1] = (value >> 8) & 0xFF
    buffer[index+2] = (value >> 16) & 0xFF
    buffer[index+3] = (value >> 24) & 0xFF

def setUint16LittleEndian(buffer, index, value) :
    '''
    set Uint16 Little Endian.

    Parameters
    ----------
    buffer : bytearray
        data array.
    index : int
        starting index.
    value : int
        value to be set.
    '''
    buffer[index] = value & 0xFF
    buffer[index+1] = (value >> 8) & 0xFF

def setInt16LittleEndian(buffer, index, value) :
    '''
    set Int16 Little Endian.

    Parameters
    ----------
    buffer : bytearray
        data array.
    index : int
        starting index.
    value : int
        value to be set.
    '''
    buffer[index] = value & 0xFF
    buffer[index+1] = (value >> 8) & 0xFF

def getUint32LittleEndian(array, index) :
    '''
    get Uint32 Little Endian.

    Parameters
    ----------
    array : bytearray
        data array.
    index : int
        starting index.
    '''
    byte0 = array[index]
    byte1 = array[index+1]
    byte2 = array[index+2]
    byte3 = array[index+3]
    value = (byte3 << 24) | (byte2 << 16) | (byte1 << 8) | byte0
    return value

def getUint16LittleEndian(array, index) :
    '''
    get Uint16 Little Endian.

    Parameters
    ----------
    array : bytearray
        data array.
    index : int
        starting index.
    '''
    byte0 = array[index]
    byte1 = array[index+1]
    value = (byte1 << 8) | byte0
    return value

def getValueLsb(value) :
    '''
    get Lsb value.

    Parameters
    ----------
    value : int
        masked value.
    '''
    return (value & 0xFFFF)

def getValueMsb(value) :
    '''
    get Msb value.

    Parameters
    ----------
    value : int
        masked value.
    '''
    return (value >> 16)
